//-----------------------------------------------------------------------
//  This file is part of the Microsoft Robotics Studio Code Samples.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//  $File: Atom.cs $ $Revision: 1 $
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using System.Web;

using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.Core.DsspHttp;
using Microsoft.Dss.Core.DsspHttpUtilities;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;

using submgr = Microsoft.Dss.Services.SubscriptionManager;
using co = Microsoft.Dss.Services.ConsoleOutput.Proxy;

using W3C.Soap;

namespace Microsoft.Robotics.Services.Syndication.Atom
{
    /// <summary>
    /// The Atom Service Implements an Atom Syndication Generator Service.
    /// </summary>
    [DisplayName("Atom Syndication Generator")]
    [Description("Provides support creating and updating web resources (e.g. blogs) using the Atom Syndication format.")]
    [Contract(Contract.Identifier)]
    public class AtomService : DsspServiceBase
    {
        #region EmbeddedResources
        // Embedded XML resource providing alternative UI for feed
        [EmbeddedResource("Microsoft.Robotics.Services.Syndication.Atom.Atom.xml")]
        string _alternateWebPage = null;
        #endregion

        #region ServicePartners
        // Subscription manager for handling subscriptions for changes to feed
        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        private submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        // Service instance state containing the feed and entry information. Feed is persisted to file.
        [InitialStatePartner(Optional = true, ServiceUri = ServicePaths.Store + @"/AtomFeed.xml")]
        private AtomFeed _state = null;

        // Optional Console Output service that we subscribe to
        [Partner("ConsoleOutput", Optional = true, Contract = co.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        co.ConsoleOutputPort _consoleoutputPort = new co.ConsoleOutputPort();
        co.ConsoleOutputPort _consoleoutputNotificationPort = null;
        #endregion

        #region HttpHelpers
        // HTTP utility port for various HTTP helper functions
        DsspHttpUtilitiesPort _utilitiesPort;
        #endregion

        // Max queue size for how many atom entries we keep
        private const int MaxSize = 200;

        #region ServiceOperationPort
        [ServicePort("/atom")]
        private AtomOperations _mainPort = new AtomOperations();
        #endregion

        #region ServiceConstructor
        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public AtomService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }
        #endregion

        #region ServiceStartup
        /// <summary>
        /// Service start
        /// </summary>
        protected override void Start()
        {
            // Set up our HTTP helper
            _utilitiesPort = DsspHttpUtilitiesService.Create(Environment);

            // First check whether we had initial state or have to create it
            if (_state == null)
            {
                _state = new AtomFeed("DSS Atom Syndication Feed");
                _state.LinkList.Add(new AtomLink(_alternateWebPage, AtomLinkRelation.Alternate));
            }

            // Continue with remaining of the start processing
            base.Start();

            // Check whether we should subscribe to console output service
            if (_consoleoutputPort != null)
            {
                // Start listening for event notifications
                _consoleoutputNotificationPort = new co.ConsoleOutputPort();
                Activate<ITask>(
                    Arbiter.Receive<co.Insert>(true, _consoleoutputNotificationPort, ConsoleOutputNotificationHandler));

                // Subscribe to console output service for warning and error messages
                co.SubscribeRequest subReq = new co.SubscribeRequest();
                subReq.Level = new XmlQualifiedName("Warning", co.Contract.Identifier);
                _consoleoutputPort.Subscribe(subReq, _consoleoutputNotificationPort);
            }
        }
        #endregion

        #region Notification Handlers
        private void ConsoleOutputNotificationHandler(co.Insert insert)
        {
            // Extract information from console output notification
            string title = 
                String.Format("Console Output {0}", 
                    ((insert.Body != null && insert.Body.Level != null) ? insert.Body.Level.Name : "Not labeled"));
            string subject = (insert.Body != null && !string.IsNullOrEmpty(insert.Body.Subject)) ?
                insert.Body.Subject : "No subject";

            // Create atom entry and insert it into our feed
            AtomContent content = new AtomContent(subject, "text");
            AtomEntry entry = new AtomEntry(title, content);
            InsertEntry insReq = new InsertEntry();
            insReq.Body = entry;
            _mainPort.Post(insReq);
        }        
        #endregion

        #region ProtocolHandlers
        #region DSSP
        #region DSSP_GET
        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get">The GET request</param>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }
        #endregion

        #region DSSP_SUBSCRIBE
        /// <summary>
        /// Subscription handler
        /// </summary>
        /// <param name="subscribe">The SUBSCRIBE request</param>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            yield return Arbiter.Choice(SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    base.SendNotification<Replace>(_submgrPort, subscribe.Body.Subscriber, _state);
                },
                delegate(Exception failure)
                {
                    LogError(null, "Unable to subscribe", failure);
                }
            );
            yield break;
        }
        #endregion

        #region DSSP_INSERT
        /// <summary>
        /// InsertEntry Handler
        /// </summary>
        /// <param name="insert">The INSERT request</param>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> InsertEntryHandler(InsertEntry insert)
        {
            AtomEntry entryInsert = insert.Body;

            // Check uniqueness of entry ID
            if (_state.EntryList.Exists(entryInsert.Equals))
            {
                insert.ResponsePort.Post(
                    Fault.FromCodeSubcodeReason(
                        FaultCodes.Sender,
                        DsspFaultCodes.DuplicateEntry,
                        entryInsert.id));
                yield break;
            }

            // Add message to queue and remove old entries if no space
            if (_state.EntryList.Count > MaxSize)
            {
                _state.EntryList.RemoveRange(0, _state.EntryList.Count - MaxSize);
            }
            _state.EntryList.Add(entryInsert);
            _state.updated = DateTime.UtcNow.ToString("o");

            // Persist new state. We only check for Fault responses
            yield return Arbiter.Choice(base.SaveState(_state),
                EmptyHandler,
                delegate(W3C.Soap.Fault fault)
                {
                    LogError(null, "Unable to store state", fault);
                }
            );

            // Send notification
            base.SendNotification(_submgrPort, insert);

            // Send response
            insert.ResponsePort.Post(DefaultInsertResponseType.Instance);
            yield break;
        }
        #endregion

        #region DSSP_UPDATE_ENTRY
        /// <summary>
        /// UpdateEntry Handler
        /// </summary>
        /// <param name="update">The UPDATE request</param>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> UpdateEntryHandler(UpdateEntry update)
        {
            AtomEntry entryUpdate = update.Body;

            // Search for atom entry in our state
            AtomEntry entry = _state.EntryList.Find(entryUpdate.Equals);
            if (entry != null)
            {
                if (entryUpdate.author != null)
                {
                    entry.author = entryUpdate.author;
                }

                if (entryUpdate.content != null)
                {
                    entry.content = entryUpdate.content;
                }

                if (!String.IsNullOrEmpty(entryUpdate.title))
                {
                    entry.title = entryUpdate.title;
                }

                if (!String.IsNullOrEmpty(entryUpdate.updated))
                {
                    entry.updated = entryUpdate.updated;
                }

                // Persist new state. Only look for faults
                yield return Arbiter.Choice(base.SaveState(_state),
                    EmptyHandler,
                    delegate(W3C.Soap.Fault fault)
                    {
                        LogError(null, "Unable to store state", fault);
                    }
                );

                // Send notification
                base.SendNotification(_submgrPort, update);

                // Send response
                update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                yield break;
            }

            update.ResponsePort.Post(
                Fault.FromCodeSubcodeReason(
                    FaultCodes.Sender,
                    DsspFaultCodes.UnknownEntry,
                    entryUpdate.id));
            yield break;
        }
        #endregion

        #region DSSP_DELETE
        /// <summary>
        /// DeleteEntry Handler
        /// </summary>
        /// <param name="delete">The DELETE request</param>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> DeleteEntryHandler(DeleteEntry delete)
        {
            AtomEntry entryDelete = delete.Body;

            // Search for atom entry in our state
            AtomEntry entry = _state.EntryList.Find(entryDelete.Equals);
            if (entry != null)
            {
                _state.EntryList.Remove(entry);
                _state.updated = DateTime.UtcNow.ToString("o");

                // Persist new state
                yield return Arbiter.Choice(base.SaveState(_state),
                    EmptyHandler,
                    delegate(W3C.Soap.Fault fault)
                    {
                        LogError(null, "Unable to store state", fault);
                    }
                );

                // Send notification
                base.SendNotification(_submgrPort, delete);

                // Send response
                delete.ResponsePort.Post(DefaultDeleteResponseType.Instance);
                yield break;
            }

            delete.ResponsePort.Post(
                Fault.FromCodeSubcodeReason(
                    FaultCodes.Sender,
                    DsspFaultCodes.UnknownEntry,
                    entryDelete.id));
            yield break;
        }
        #endregion

        #region DSSP_UPDATE_FEED
        /// <summary>
        /// Update Feed Handler
        /// </summary>
        /// <param name="update">The UPDATE feed request</param>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> UpdateFeedHandler(UpdateFeed update)
        {
            AtomFeed feedUpdate = update.Body;

            if (feedUpdate.author != null)
            {
                _state.author = feedUpdate.author;
            }

            if (!String.IsNullOrEmpty(feedUpdate.title))
            {
                _state.title = feedUpdate.title;
            }

            if (!String.IsNullOrEmpty(feedUpdate.updated))
            {
                _state.updated = feedUpdate.updated;
            }

            // Persist new state
            yield return Arbiter.Choice(base.SaveState(_state),
                EmptyHandler,
                delegate(W3C.Soap.Fault fault)
                {
                    LogError(null, "Unable to store state", fault);
                }
            );

            // Send notification
            base.SendNotification(_submgrPort, update);

            // Send response
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }
        #endregion
        #endregion

        #region HTTP
        #region HTTP_GET
        /// <summary>
        /// HTTP Get Handler
        /// </summary>
        /// <param name="httpGet">The HTTP GET request</param>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> HttpGetHandler(HttpGet httpGet)
        {
            httpGet.ResponsePort.Post(new HttpResponseType(System.Net.HttpStatusCode.OK, _state));
            yield break;
        }
        #endregion

        #region HTTP_QUERY
        /// <summary>
        /// HTTP Query Handler
        /// </summary>
        /// <param name="httpQuery">The HTTP GET request with a query request URI.</param>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> HttpQueryHandler(HttpQuery httpQuery)
        {
            // Getting ready to generate query response
            AtomFeed feedCopy = new AtomFeed(_state);

            // Get query parameters
            NameValueCollection nvCol = httpQuery.Body.Query;
            string q = nvCol.Get("q");

            // Find query matches in title and contents
            // An entry is only considered a match if all keywords match (and)
            if (!String.IsNullOrEmpty(q))
            {
                string[] kwList = AtomUtilities.QueryTokenizer(q);
                feedCopy.EntryList = _state.EntryList.FindAll(
                    delegate(AtomEntry entry)
                    {
                        int found = 0;
                        foreach (string kw in kwList)
                        {
                            if ((!string.IsNullOrEmpty(entry.title) &&
                                entry.title.IndexOf(kw, StringComparison.InvariantCultureIgnoreCase) >= 0) ||
                                (entry.content != null && !string.IsNullOrEmpty(entry.content.content) &&
                                entry.content.content.IndexOf(kw, StringComparison.InvariantCultureIgnoreCase) >= 0))
                            {
                                found++;
                            }
                        }
                        return (found >= kwList.Length);
                    }
                );
            }
            else
            {
                // No query keywords
                feedCopy.EntryList = new List<AtomEntry>();
            }

            // Send response
            httpQuery.ResponsePort.Post(new HttpResponseType(System.Net.HttpStatusCode.OK, feedCopy));
            yield break;
        }
        #endregion

        #region HTTP_POST
        /// <summary>
        /// HTTP Post Handler
        /// </summary>
        /// <param name="httpPost">The HTTP POST request</param>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> HttpPostHandler(HttpPost httpPost)
        {
            Fault fault = null;
            NameValueCollection nvCol = null;

            ReadFormData readForm = new ReadFormData(httpPost.Body.Context);
            _utilitiesPort.Post(readForm);

            yield return Arbiter.Choice(
                readForm.ResultPort,
                delegate(NameValueCollection col)
                {
                    nvCol = col;
                },
                delegate(Exception e)
                {
                    fault = Fault.FromException(e);
                    LogError(null, "Error processing form data", fault);
                }
            );

            if (fault != null)
            {
                httpPost.ResponsePort.Post(fault);
                yield break;
            }

            string title = nvCol["title"];
            string content = nvCol["content"];
            string value = nvCol["value"];
            string type = nvCol["type"] ?? AtomType.Text;

            // If incomplete submission
            if (String.IsNullOrEmpty(title) || String.IsNullOrEmpty(content) ||
                (!AtomType.IsHtml(type) && !AtomType.IsText(type)))
            {
                httpPost.Body.Context.Response.StatusCode = (int)HttpStatusCode.NoContent;
                Utilities.HttpClose(httpPost.Body.Context);
                yield break;
            }

            // Create new content by checking whether it is inline or not
            AtomContent entryContent = new AtomContent();
            if (!String.IsNullOrEmpty(value) &&
                String.Compare(value, "inline", StringComparison.InvariantCultureIgnoreCase) == 0)
            {
                entryContent.type = AtomType.Canonicalize(type);
                entryContent.content = content;
            }
            else
            {
                entryContent.src = content;
            }

            // Create entry and fill it in with our new entry
            AtomEntry entry = new AtomEntry(title, entryContent);
            InsertEntry insert = new InsertEntry();
            insert.Body = entry;

            // Call insert handler and wait for it to complete using the ExecuteToCompletion arbiter
            // Note, this works even when this handler *and* the InsertEntryHandler are marked as 
            // exclusive on main port.
            yield return Arbiter.ExecuteToCompletion(TaskQueue,
                new IterativeTask<InsertEntry>(insert, InsertEntryHandler)); 

            // Send response and close HTTP stream
            httpPost.ResponsePort.Post(new HttpResponseType(System.Net.HttpStatusCode.OK, _state));
            yield break;
        }
        #endregion
        #endregion
        #endregion
    }
}
