//-----------------------------------------------------------------------
//  This file is part of the Microsoft Robotics Studio Code Samples.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//  $File: WinMessage.cs $ $Revision: 1 $
//-----------------------------------------------------------------------

using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml;

using System.Runtime.InteropServices;
using System.Diagnostics;

using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Imaging;
using Microsoft.Ccr.Adapters.WinForms;

using submgr = Microsoft.Dss.Services.SubscriptionManager;
using winmessage = Microsoft.Robotics.Services.Sample.WinMessage;
using W3C.Soap;


namespace Microsoft.Robotics.Services.Sample.WinMessage
{
    
    /// <summary>
    /// Implementation class for WinMessage
    /// </summary>
    [DisplayName("Windows Messages Sample")]
    [Description("Demonstrates how to send and receive Windows messages.")]
    [Contract(Contract.Identifier)]
    public class WinMessageService : DsspServiceBase
    {
        /// <summary>
        /// _state
        /// </summary>
        [InitialStatePartner(Optional=true, ServiceUri="WinMessage.config.xml")]
        private WinMessageState _state = new WinMessageState();

        /// <summary>
        /// _main Port
        /// </summary>
        [ServicePort("/winmessage", AllowMultipleInstances=true)]
        private WinMessageOperations _mainPort = new WinMessageOperations();

        /// <summary>
        /// Internal port to update private state.
        /// </summary>
        private WinMessageOperations _internalPort = new WinMessageOperations();

        private Port<bool> _formInitializedPort = new Port<bool>();
        private WinMsgForm _form;


        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public WinMessageService(DsspServiceCreationPort creationPort) : 
                base(creationPort)
        {
        }

        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        private submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
            if (_state == null)
            {
                _state = new WinMessageState();
                SaveState(_state);
            }

            // Publish the service to the local Node Directory
            DirectoryInsert();

            // Start the Message Window
            WinFormsServicePort.Post(new RunForm(CreateVisionForm));

            // Finalize the service initialization only 
            // after the form finishes loading
            Activate(Arbiter.Receive<bool>(false, _formInitializedPort, FinalizeStartup));
        }

        Form CreateVisionForm()
        {
            _form = new WinMsgForm(_mainPort, _internalPort, _state, _formInitializedPort);
            ConfigureForm();
            return _form;
        }

        private void FinalizeStartup(bool ok)
        {
            ActivateDsspOperationHandlers();
            MainPortInterleave.CombineWith(
                Arbiter.Interleave(
                    new TeardownReceiverGroup(
                        Arbiter.Receive<Shutdown>(false, CleanupPort, delegate(Shutdown done) { CleanupPort.Post(done); })
                    ),
                    new ExclusiveReceiverGroup(
                        Arbiter.Receive<Configure>(true, _internalPort, FormConfigureHandler)
                    ),
                    new ConcurrentReceiverGroup(
                        Arbiter.Receive<ReceiveMessage>(true, _internalPort, ReceiveMessageHandler)
                    )
            ));
        }
        /// <summary>
        /// Get Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Configure Handler
        /// </summary>
        /// <param name="configure"></param>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual void ReplaceHandler(Configure configure)
        {
            _state = configure.Body;
            ConfigureForm();
            configure.ResponsePort.Post(DefaultReplaceResponseType.Instance);
        }

        
        /// <summary>
        /// Update service based on configuration changes made to the form.
        /// </summary>
        /// <param name="configure"></param>
        private void FormConfigureHandler(Configure configure)
        {
            if (_state.ProcessName != configure.Body.ProcessName
                || _state.ShowWindow != configure.Body.ShowWindow)
            {
                _state.ProcessName = configure.Body.ProcessName;
                _state.ShowWindow = configure.Body.ShowWindow;
                SaveState(_state);
            }
        }

        /// <summary>
        /// Configure Form
        /// </summary>
        private void ConfigureForm()
        {
            if (_form != null)
                _form.SetConfiguration(_state);
        }

        /// <summary>
        /// Subscribe Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            LogInfo(LogGroups.Console, "Subscribe request from: " + subscribe.Body.Subscriber);
            SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort);
            yield break;
        }

        /// <summary>
        /// ReceiveMessage External Handler for throwing exception
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> ReceiveMessageExternal(ReceiveMessage update)
        {
            update.ResponsePort.Post(Fault.FromException(new InvalidOperationException("ReceiveWinMsg is only valid for notifications")));
            yield break;
        }

        /// <summary>
        /// ReceiveMessage Handler
        /// </summary>
        public void ReceiveMessageHandler(ReceiveMessage update)
        {
            const int WM_COPYDATA = 0x004A;

            if(update.Body.Msg == WM_COPYDATA)
                LogVerbose(LogGroups.Console, string.Format("Got WindowMessage: WM_COPYDATA, data {1}, wParam {2}", update.Body.Msg, update.Body.StringData, update.Body.WParam));
            else
                LogVerbose(LogGroups.Console, string.Format("Got WindowMessage: Msg {0}, wParam {1}, lParam {2}", update.Body.Msg, update.Body.WParam, update.Body.LParam));

            base.SendNotification(_submgrPort, update);
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        /// <summary>
        /// SendMessage Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SendMessageHandler(SendMessage update)
        {
            const int HWND_BROADCAST = 0xffff;
            IntPtr HandleOfToProcess = IntPtr.Zero;

            if (update.Body.ProcessName == "HWND_BROADCAST")
            {
                HandleOfToProcess = (IntPtr)HWND_BROADCAST;
            }
            else
                HandleOfToProcess = FindProcessByName(update.Body.ProcessName);

            NativeMethods.SendMessage(HandleOfToProcess, update.Body.Msg, (IntPtr)update.Body.WParam, (IntPtr)update.Body.LParam);
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// SendMessageString Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SendMessageStringHandler(SendMessageString update)
        {
            const int HWND_BROADCAST = 0xffff;
            const int WM_COPYDATA = 0x004A;
            IntPtr HandleOfToProcess = IntPtr.Zero;

            if (update.Body.ProcessName == "HWND_BROADCAST")
            {
                HandleOfToProcess = (IntPtr)HWND_BROADCAST;
            }
            else
                HandleOfToProcess = FindProcessByName(update.Body.ProcessName);

            NativeMethods.SendMessage(HandleOfToProcess, WM_COPYDATA, IntPtr.Zero, new CopyDataStruct(update.Body.StringData));

            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// PlaySound Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public void PlaySoundHandler(PlaySound update)
        {
            if (update.Body.Flags == PlaySoundFlags.SND_ASYNC)
                update.Body.Flags |= PlaySoundFlags.SND_FILENAME;

            int result = NativeMethods.PlaySound(update.Body.FileName, IntPtr.Zero, (int)update.Body.Flags);
            if (result == 1)
                update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            else
                update.ResponsePort.Post(Fault.FromException(new ArgumentException(string.Format("Failed to play: {0}", update.Body.FileName))));
        }

        IntPtr FindProcessByName(string processName)
        {
            IntPtr HandleOfToProcess = IntPtr.Zero;

            Process[] p = Process.GetProcesses();
            foreach (Process p1 in p)
            {
                if (p1.ProcessName == processName)
                {
                    HandleOfToProcess = p1.MainWindowHandle;
                    //LogInfo(LogGroups.Console, string.Format("found Process {0}", processName));
                    break;
                }
            }

            if (HandleOfToProcess == IntPtr.Zero)
            {
                HandleOfToProcess = NativeMethods.FindWindow(null, processName);
                //if (HandleOfToProcess != IntPtr.Zero) LogInfo(LogGroups.Console, string.Format("found WindowName {0}", processName));
            }

            return HandleOfToProcess;
        }
    }

    /// <summary> 
    /// Structure required to be sent with the WM_COPYDATA message 
    /// </summary> 
    [StructLayout(LayoutKind.Sequential)]
    class CopyDataStruct
    {
        public int dwData = 1;//32 bit int to passed.
        public int cbData = 0;//length of string. Will be one greater because of null termination. 
        public string lpData;//string to be passed. 

        public CopyDataStruct()
        { }
        public CopyDataStruct(string Data)
        {
            lpData = Data + "\0";   //add null termination 
            cbData = lpData.Length + 1; //length includes null chr so will be one greater 
        }
    } 

    internal static class NativeMethods
    {
        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        internal static extern IntPtr SendMessage(IntPtr hwnd, int wMsg, IntPtr wParam, CopyDataStruct lParam);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        internal static extern IntPtr SendMessage(IntPtr hwnd, int wMsg, IntPtr wParam, IntPtr lParam);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        internal static extern IntPtr WindowFromPoint(int xPoint, int yPoint);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        internal static extern IntPtr FindWindow(string lpClassName, string lpWindowName);

        [DllImport("winmm.dll", CharSet = CharSet.Auto)]
        internal static extern int PlaySound(string filename, IntPtr hModule, int dwFlags);
    }
}
