//-----------------------------------------------------------------------
//  This file is part of the Microsoft Robotics Studio Code Samples.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//  $File: XInputGamepad.cs $ $Revision: 1 $
//-----------------------------------------------------------------------
using System;
using System.Net;
using System.Collections.Generic;
using System.ComponentModel;
using System.Security.Permissions;

using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.Core.DsspHttp;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using sm = Microsoft.Dss.Services.SubscriptionManager;

using xinput = Microsoft.Xna.Framework.Input;
using xna = Microsoft.Xna.Framework;
using W3C.Soap;


namespace Microsoft.Robotics.Services.Sample.XInputGamepad
{
    
    [DisplayName("XInput Controller")]
    [Description("Provides access to an Xbox 360 controller such as a gamepad.")]
    [Contract(Contract.Identifier)]
    public class XInputGamepadService : DsspServiceBase
    {
        [EmbeddedResource("Microsoft.Robotics.Services.Sample.XInputGamepad.GamePad.xslt")]
        string _transform = null;
        
        [InitialStatePartner(Optional = true)]
        private XInputGamepadState _state;
        
        [ServicePort("/xinputgamepad", AllowMultipleInstances=true)]
        private XInputGamepadOperations _mainPort = new XInputGamepadOperations();

        [Partner("SubMgr", Contract = sm.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways, Optional = false)]
        sm.SubscriptionManagerPort _submgr = new sm.SubscriptionManagerPort();

        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public XInputGamepadService(DsspServiceCreationPort creationPort) : 
                base(creationPort)
        {
			
        }
        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
            if (_state == null)
            {
                _state = new XInputGamepadState();
            }

            // Listen on the main port for requests and call the appropriate handler.
            ActivateDsspOperationHandlers();

            // Publish the service to the local service Directory
            DirectoryInsert();

            // start the timer
            Spawn(DateTime.Now, TimerHandler);
        }
        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Http Get Handler.  Needed for XSLT transform
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> HttpGetHandler(HttpGet httpGet)
        {
            HttpResponseType rsp = new HttpResponseType(HttpStatusCode.OK, _state, _transform);
            httpGet.ResponsePort.Post(rsp);
            yield break;
        }

        void TimerHandler(DateTime signal)
        {
            _mainPort.Post(new Poll(new PollRequest()));
            Activate(
                Arbiter.Receive(false, TimeoutPort(50), TimerHandler)
            );
        }
        
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> PollHandler(Poll poll)
        {
            DateTime timeStamp = DateTime.Now;
            xinput.GamePadState gamepad = _state.GetState();

            if (_state.Update(timeStamp, gamepad))
            {
                if (_state.Controller.Update(gamepad))
                {
                    SendNotification<ControllerChanged>(_submgr, _state.Controller);
                }

                if (_state.DPad.Update(timeStamp, gamepad.DPad))
                {
                    SendNotification<DPadChanged>(_submgr, _state.DPad);
                }

                if (_state.Buttons.Update(timeStamp, gamepad.Buttons))
                {
                    SendNotification<ButtonsChanged>(_submgr, _state.Buttons);
                }

                if (_state.Triggers.Update(timeStamp, gamepad.Triggers))
                {
                    SendNotification<TriggersChanged>(_submgr, _state.Triggers);
                }

                if (_state.Thumbsticks.Update(timeStamp, gamepad.ThumbSticks))
                {
                    SendNotification<ThumbsticksChanged>(_submgr, _state.Thumbsticks);
                }
            }

            poll.ResponsePort.Post(DefaultSubmitResponseType.Instance);
            yield break;
        }
        /// <summary>
        /// Replace Handler
        /// </summary>
        /// <param name="replace"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> ReplaceHandler(Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);

            SendNotification(_submgr, replace);
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            SubscribeRequestType request = subscribe.Body;

            yield return Arbiter.Choice(SubscribeHelper(_submgr, request, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    SendNotification<Replace>(_submgr, request.Subscriber, _state);
                },
                delegate(Exception failure){}
            );
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> ControllerChangedHandler(ControllerChanged controllerChanged)
        {
            Controller controller = controllerChanged.Body;

            if (controller.IsValid)
            {
                _state.Controller.TimeStamp = DateTime.Now;
                _state.Controller.PlayerIndex = controller.PlayerIndex;
                _state.Controller.Update(_state.GetState());

                SendNotification<ControllerChanged>(_submgr, _state.Controller);
            }
            else
            {
                controllerChanged.ResponsePort.Post(
                    Fault.FromCodeSubcodeReason(
                        FaultCodes.Receiver,
                        DsspFaultCodes.UnknownEntry,
                        "Bad Controller"
                    )
                );
            }

            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> SetVibrationHandler(SetVibration setVibration)
        {
            _state.SetVibration(setVibration.Body);
            setVibration.ResponsePort.Post(DefaultUpdateResponseType.Instance);

            SendNotification(_submgr, setVibration);
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> QueryControllersHandler(QueryControllers queryControllers)
        {
            QueryControllersResponse response = new QueryControllersResponse();

            for (PlayerIndex player = PlayerIndex.One; player <= PlayerIndex.Four; player++)
            {
                xinput.GamePadCapabilities caps = xinput.GamePad.GetCapabilities((xna.PlayerIndex)player);

                if (caps.IsConnected)
                {
                    ControllerCaps controller = new ControllerCaps();
                    controller.PlayerIndex = player;
                    controller.IsConnected = true;
                    controller.ControllerType = (ControllerType)caps.GamePadType;

                    response.Controllers.Add(controller);
                }
            }

            queryControllers.ResponsePort.Post(response);

            yield break;
        }
    }
}
